/*
 * Copyright (C) 2000,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 *
 * $Revision: 1.2 $
 * $Name: PublicRelease-21Jan00-B $
 * $Log: sockio.c,v $
 * Revision 1.2  2000/01/21 21:28:16  wes
 * Added negotiation code so that the client tells the server how big
 * the TCP window will be. The negotiation process involves opening a
 * socket, sending over the config info, closing the reopening another
 * socket after the TCP window size has been set. This allow for
 * quick testing using varying TCP window sizes w/o the need for recompiling
 * both client & server.
 *
 * Revision 1.1.1.1  2000/01/05 21:49:36  wes
 * Initial checkin
 *
 */
 
#include "slam.h"

int Debug_level = 0;	/* DAVIDR */
/*
 * all routines in this file are thread/MPI safe.
 */



int
write_n(int fd,
       const void *src,
       int nbytesToWrite)
{
    /*
     * return # of bytes written or -1 on error
     */
    size_t nleft;
    ssize_t nwritten;
    const char *ptr;
    
    struct timeval t1, t2;
    double dt1,dt2,dt;

    ptr = src;
    nleft = nbytesToWrite;
    
    gettimeofday(&t1,NULL);
    
    while (nleft > 0)
    {
	if ((nwritten = write(fd,ptr,nleft)) <= 0)
	{
	    if (errno == EINTR) /* pthreads: each thread gets it's own errno */
	    {
		fprintf(stderr," write_n: caught EINTR \n");
		nwritten = 0;	/* call write() again */
	    }
	    else
	    {
		perror(" write_n error");
		return(-1);	/* error */
	    }
	}
	if (Debug_level == DEBUG_ALL)
	  fprintf(stderr,"write_n: wrote %d bytes \n",nwritten);
	nleft -= nwritten;
	ptr += nwritten;
    }

    gettimeofday(&t2,NULL);
    dt1 = t1.tv_sec + (double) t1.tv_usec/1000000.0;
    dt2 = t2.tv_sec + (double) t2.tv_usec/1000000.0;
    dt = dt2 - dt1;		/* in seconds */
    if (Debug_level == DEBUG_ALL)
      fprintf(stderr, "*** write_n (%d bytes) Mb/s %lf \n",nbytesToWrite,
	    nbytesToWrite/dt/128/1000);
    
    return(nbytesToWrite);
}

int
read_n(int fd,
      void *dst,
      int nbytesExpected,
      int myID)
{
    /*
     * return # of bytes read or -1 on error
     */
    size_t nleft;
    ssize_t nread;
    char *ptr;
    
    ptr = dst;
    nleft = nbytesExpected;
    while (nleft > 0)
    {
	if ((nread=read(fd,ptr,nleft)) < 0)
	{
	    if (errno == EINTR) /* pthreads: each thread gets it's own errno */
	    {
		nread = 0;	/* call read again */
		fprintf(stderr," EINTR signal caught \n");
	    }
	    else
	    {
		fprintf(stderr," read_n: nread < 0 \n");
		return(-1);
	    }
	}
	else if (nread == 0)
	{
	    fprintf(stderr," read_n: nread==0 \n");
	    break;		/* EOF */
	}
#if 0
	fprintf(stderr,"%d read_n: read %d bytes \n",myID,nread);
#endif
	nleft -= nread;
	ptr += nread;
    }

    if (Debug_level == DEBUG_ALL)
      fprintf(stderr, "--- read_n: PE/thread %d, read %d bytes \n",myID,(int)(ptr-(char *)dst));

    return(nbytesExpected-nleft);
}

int
readline(int fd,
	 void *dst,
	 int maxBytes)
{
    /*
     * return # of bytes read or -1 on error
     * this routine is a pig: does read() 1 char at a time.
     */

    ssize_t t, rc;
    char c, *ptr;
    int n;

    ptr = dst;

#if 0
    /* new stuff, doesn't work right. */
    for (n=0; n<maxBytes;n++)
    {
	if ((rc = read(fd,&c,1)) == 1)
	{
	    *ptr++ = c;
	    if (c == '\n')	/* poss read one more, the '\0' */
	    {
		rc = read(fd,&c,1);
		break;
	    }
	}
	else if (rc == 0)
        {
	    if (n == 0)
		return(0);	/* eof, no data read */
	    else
		break;		/* eof, some data was read */
	}
	else
	{
	    if (errno != EINTR)
		return(-1);
	}
    }
#endif

    for (n=1; n<maxBytes; n++)
    {
    again:			/* argh! */
	if ( (rc = read(fd, &c, 1)) == 1)
	{
	    *ptr++ = c;
	    if (c == '\n')
		break;
	}
	else if (rc == 0)
	{
	    if (n == 1)
		return(0);	/* eof, no data read */
	    else
		break;	/* eof, some data was read */
	}
	else
	{
	    if (errno == EINTR) /* pthreads: each thread gets it's own errno */
		goto again;
	    return(-1);
	}
    }

    *ptr = 0;			/* null terminate like fgets() */
    return(n);
}

int
winshiftFromBufsize(int bufsize)
{
    int a,i;
    int winshift;

    a = bufsize / 65536;

    winshift = 0;
    for (i=0; a > 0; i++, winshift++)
    {
        a = a >> 1;
    }
    if (i > 0)
        winshift -= 1;

    return(winshift);
}


void
setTCPWindowSize(int connfd,
		 int reqTCPWindowSize)
{
    /*
     * use this routine to set the TCP window size. this is accomplished
     * by setting the socket's send & recv buffer sizes. on Unicos, we
     * use the TCP winshift parameter rather than the socket buffer
     * sizes to manipulate the TCP window size. Unicos uses a
     * TCP window size that is an integral multiple of 64K in size.
     */
    int bufsize=reqTCPWindowSize;
#ifdef UNICOS
    int winshift;

    winshift = winshiftFromBufsize(bufsize);

    if (setsockopt(connfd, IPPROTO_TCP, TCP_WINSHIFT,
		   (char *) &winshift, sizeof (winshift)))
	perror (" setsockopt winshift client-side fails. ");
#else
    if (setsockopt(connfd, SOL_SOCKET, SO_SNDBUF, (void *)&bufsize,
		   sizeof(int)) < 0)
	perror("setsockopt SO_SNDBUF client-side");
	
    if (setsockopt(connfd, SOL_SOCKET, SO_RCVBUF, (void *)&bufsize,
		   sizeof(int)) < 0)
	perror("setsockopt SO_RCVBUF client-side");
#endif
}


/*
 * connection management
 */


int
connectToServerSettingTCPWindow(char *serverHostString,
				int haveHostname,
				int myBlockSize)
{
    struct sockaddr_in cliaddr, servaddr;
    int outfd;
    struct hostent *h;
    
    outfd = socket(AF_INET, SOCK_STREAM, 0);

    setTCPWindowSize(outfd,myBlockSize);

    memset(&servaddr,0,sizeof(struct sockaddr_in));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(SLAM_SERVER_PORT);

    if (haveHostname == 1)
    {
	h = gethostbyname(serverHostString);
	memcpy(&((servaddr.sin_addr)),h->h_addr,4);
    }
    else
	servaddr.sin_addr.s_addr = inet_addr(serverHostString);

    sleep(4);
    connect(outfd, (struct sockaddr *)&servaddr,
	    sizeof(servaddr));

#ifdef DEBUG
    fprintf(stderr," connected to %s on port %d. \n",
	    inet_ntoa(servaddr.sin_addr), ntohs(servaddr.sin_port));
#endif

    return(outfd);
}
/*
 * establish a connection to a server host.  return a file descriptor.
 * input:
 * 1. serverHostString. a character string containing either the hostname
 *        (eg, "gonzo.lbl.gov" or a dotted quad, eg "128.31.241.11")
 * 2. haveHostname - an int (really a boolean). a value of zero means
 *        the string is a dotted quad, a value of 1 means the string
 *        has a hostname and this routine will attempt to find the
 *        dotted quad from the hostname.
 */
int
connectToServer(char *serverHostString,
		int haveHostname,
		int myBlockSize)
{
    struct sockaddr_in cliaddr, servaddr;
    int outfd;
    struct hostent *h;
    
    outfd = socket(AF_INET, SOCK_STREAM, 0);

    /* this routine doesn't try to set the TCP window size */

    memset(&servaddr,0,sizeof(struct sockaddr_in));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(SLAM_SERVER_PORT);

    if (haveHostname == 1)
    {
	h = gethostbyname(serverHostString);
	memcpy(&((servaddr.sin_addr)),h->h_addr,4);
    }
    else
	servaddr.sin_addr.s_addr = inet_addr(serverHostString);

    connect(outfd, (struct sockaddr *)&servaddr,
	    sizeof(servaddr));

#ifdef DEBUG
    fprintf(stderr," connected to %s on port %d. \n",
	    inet_ntoa(servaddr.sin_addr), ntohs(servaddr.sin_port));
#endif

    return(outfd);

}


int
setupServerSocket(int serverListenToPortNum,
		  int bufSize)
{
    struct sockaddr_in servaddr;
    int listenfd;
    int on=1;
    
    if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
	perror("setupServerSocket - socket creation fails");
	return(-1);
    }

    if (setsockopt(listenfd,SOL_SOCKET,SO_REUSEADDR,(void *)&on,sizeof(on)) < 0)
    {
	fprintf(stderr," error setting SO_REUSEADDR option on socket. \n");
	return(-1);
    }

#ifdef SO_REUSEPORT
    if (setsockopt(listenfd,SOL_SOCKET,SO_REUSEPORT,(void *)&on,sizeof(on)) < 0)
    {
	fprintf(stderr," error setting SO_REUSEPORT option on socket. \n");
	return(-1);
    }
#endif

    setTCPWindowSize(listenfd, bufSize);
		       
    memset(&servaddr,0,sizeof(struct sockaddr_in));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(SLAM_SERVER_PORT);

    bind(listenfd, (const struct sockaddr *)&servaddr, sizeof(servaddr));
    fprintf(stderr, "server listening:  ");
    fflush(stderr);
    listen(listenfd,8);

    fprintf(stderr," \n");

    return(listenfd);
}


